from django import forms
from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import EmailMultiAlternatives
from django.db import connection
from django.template import TemplateDoesNotExist, loader
from django.urls import reverse
from django.utils.encoding import iri_to_uri

from mailauth.backends import MailAuthBackend


class BaseLoginForm(forms.Form):
    next = forms.CharField(widget=forms.HiddenInput, required=False)

    def get_login_url(self, request, token, next=None):
        """
        Return user login URL including the access token.

        Args:
            request (django.http.request.HttpRequest): Current request.
            token (str): The user specific authentication token.
            next (str): The path the user should be forwarded to after login.

        Returns:
            str: User login URL including the access token.

        """
        protocol = 'https' if request.is_secure() else 'http'
        current_site = get_current_site(request)
        url = '{protocol}://{domain}{path}'.format(
            protocol=protocol,
            domain=current_site.domain,
            path=reverse(
                'mailauth:login-token',
                kwargs={'token': token}
            )
        )
        if next is not None:
            url += '?next=%s' % iri_to_uri(next)
        return url

    def get_token(self, user):
        """Return the access token."""
        return MailAuthBackend.get_token(user=user)

    def get_context(self, request, user):
        """
        Return the context for a message template render.

        Args:
            request (django.http.request.HttpRequest): Current request.
            user: The user requesting a login message.

        Returns:
            dict:
                A context dictionary including:

                - site
                - site_name
                - token
                - login_url
                - user

        """
        token = self.get_token(user)
        site = get_current_site(request)
        login_url = self.get_login_url(request, token, self.cleaned_data['next'])
        return {
            'site': site,
            'site_name': site.name,
            'token': token,
            'login_url': login_url,
            'user': user,
        }

    def save(self):
        """
        Send login URL to users.

        Called from the login view, if the form is valid.

        This method must be implemented by subclasses. This method
        should trigger the login url to be sent to the user.
        """
        raise NotImplementedError


class EmailLoginForm(BaseLoginForm):
    """Login form that contains only the Users email field."""

    subject_template_name = 'registration/login_subject.txt'
    email_template_name = 'registration/login_email.txt'
    html_email_template_name = 'registration/login_email.html'
    from_email = None

    def __init__(self, request, *args, **kwargs):
        self.request = request
        super(EmailLoginForm, self).__init__(*args, **kwargs)

        self.field_name = get_user_model().get_email_field_name()
        model_field = get_user_model()._meta.get_field(self.field_name)
        field = model_field.formfield()
        field.required = True

        self.fields[self.field_name] = field

    def get_users(self, email=None):
        if connection.vendor == 'postgresql':
            query = {self.field_name: email}
        else:
            query = {'%s__iexact' % self.field_name: email}
        return get_user_model()._default_manager.filter(**query).iterator()

    def save(self):
        """
        Generate and send a one-time link for the user to login.

        This method will be called from the view and passed the views request.
        """
        email = self.cleaned_data[self.field_name]
        for user in self.get_users(email):
            context = self.get_context(self.request, user)
            self.send_mail(email, context)

    def send_mail(self, to_email, context):
        """Send a django.core.mail.EmailMultiAlternatives to `to_email`."""
        subject = loader.render_to_string(self.subject_template_name, context)
        # Email subject *must not* contain newlines
        subject = ''.join(subject.splitlines())
        body = loader.render_to_string(self.email_template_name, context)

        email_message = EmailMultiAlternatives(
            subject, body, self.from_email, [to_email]
        )
        try:
            template = loader.get_template(self.html_email_template_name)
        except TemplateDoesNotExist:
            pass
        else:
            html_email = template.render(context, self.request)
            email_message.attach_alternative(html_email, 'text/html')

        email_message.send()



from django.contrib.auth.base_user import BaseUserManager
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils.crypto import get_random_string, salted_hmac
from django.utils.translation import ugettext_lazy as _

try:
    from django.contrib.postgres.fields import CIEmailField
except ImportError:
    from django.db.models import EmailField as CIEmailField


class EmailUserManager(BaseUserManager):
    use_in_migrations = True

    def _create_user(self, email, **extra_fields):
        """Create and save a user with the given email."""
        email = self.normalize_email(email)
        user = self.model(email=email, **extra_fields)
        user.save(using=self._db)
        return user

    def create_user(self, email, **extra_fields):
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)
        return self._create_user(email, **extra_fields)

    def create_superuser(self, email, **extra_fields):
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError('Superuser must have is_staff=True.')
        if extra_fields.get('is_superuser') is not True:
            raise ValueError('Superuser must have is_superuser=True.')

        return self._create_user(email, **extra_fields)


class AbstractEmailUser(AbstractUser):
    EMAIL_FIELD = 'email'
    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = []

    username = None
    password = None

    email = CIEmailField(_('email address'), unique=True, db_index=True)
    """The field is unique and case insensitive to serve as a better username."""

    session_salt = models.CharField(
        max_length=12, editable=False,
        default=get_random_string,
    )
    """Salt for the session hash replacing the password in this function."""

    def has_usable_password(self):
        return False

    objects = EmailUserManager()

    class Meta(AbstractUser.Meta):
        abstract = True

    def get_session_auth_hash(self):
        """Return an HMAC of the session_salt field."""
        key_salt = "mailauth.contrib.user.models.EmailUserManager.get_session_auth_hash"
        if not self.session_salt:
            raise ValueError("'session_salt' must be set")
        return salted_hmac(key_salt, self.session_salt).hexdigest()


delattr(AbstractEmailUser, 'password')


class EmailUser(AbstractEmailUser):
    class Meta(AbstractEmailUser.Meta):
        swappable = 'AUTH_USER_MODEL'



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.core.signing import BadSignature
from django.urls import reverse

from . import signing


class MailAuthBackend(ModelBackend):
    signer = signing.UserSigner(sep='/')

    def authenticate(self, request, token=None):
        max_age = getattr(settings, 'LOGIN_URL_TIMEOUT', 60 * 15)
        single_use = getattr(settings, 'LOGIN_TOKEN_SINGLE_USE', True)

        try:
            user = self.signer.unsign(token, max_age=max_age, single_use=single_use)
        except (get_user_model().DoesNotExist, BadSignature):
            return
        else:
            if self.user_can_authenticate(user):
                return user

    @classmethod
    def get_token(cls, user):
        return cls.signer.sign(user)

    @staticmethod
    def get_login_url(token):
        return reverse(
            'mailauth:login-token',
            kwargs={'token': token}
        )



from django.contrib.auth.views import LogoutView
from django.urls import path, re_path

from mailauth import views

app_name = 'mailauth'

urlpatterns = [
    path('login/', views.LoginView.as_view(), name='login'),
    path('login/success/', views.LoginRequestedView.as_view(), name='login-success'),
    re_path('login/(?P<token>.*)$', views.LoginTokenView.as_view(), name='login-token'),
    path('logout/', LogoutView.as_view(), name='logout'),
]



from django.contrib.auth import REDIRECT_FIELD_NAME
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mailauth.views import LoginView


class AdminLoginView(LoginView):
    template_name = 'mailauth_admin/login.html'
    site = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            **self.site.each_context(self.request),
            'title': _('Log in'),
            'app_path': self.request.get_full_path(),
            'username': self.request.user.get_username(),
        })
        if (REDIRECT_FIELD_NAME not in self.request.GET and
                REDIRECT_FIELD_NAME not in self.request.POST):
            context[REDIRECT_FIELD_NAME] = reverse(
                'admin:index', current_app=self.site.name
            )
        return context



from django.db import migrations

try:
    from django.contrib.postgres.fields import CIEmailField
except ImportError:
    CIEmailField = None
else:
    from django.contrib.postgres.operations import CITextExtension


def _operations():
    if CIEmailField:
        yield CITextExtension()
        yield migrations.AlterField(
            model_name='emailuser',
            name='email',
            field=CIEmailField(
                db_index=True, max_length=254, unique=True, verbose_name='email address'
            ),
        )
    else:
        yield migrations.RunSQL(
            sql=(
                'CREATE UNIQUE INDEX mailauth_user_emailuser_email_upper_idx'
                ' ON mailauth_user_emailuser (UPPER("email"));',
            )
            ,
            reverse_sql=(
                'DROP INDEX mailauth_user_emailuser_email_upper_idx;',
            )
        )


class Migration(migrations.Migration):
    dependencies = [
        ('mailauth_user', '0002_emailuser_session_salt'),
    ]

    operations = list(_operations())



# Generated by Django 2.2.1 on 2019-05-27 10:39

import django.utils.crypto
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mailauth_user', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailuser',
            name='session_salt',
            field=models.CharField(default=django.utils.crypto.get_random_string, editable=False, max_length=12),
        ),
    ]
